from .return_class import AbstractApiClass


class CodeEmbeddings(AbstractApiClass):
    """
        Code embeddings

        Args:
            client (ApiClient): An authenticated API Client instance
            embeddings (dict): A dictionary mapping the file name to its embeddings.
            chunkingScheme (str): The scheme used for chunking the embeddings.
    """

    def __init__(self, client, embeddings=None, chunkingScheme=None):
        super().__init__(client, None)
        self.embeddings = embeddings
        self.chunking_scheme = chunkingScheme
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'embeddings': repr(
            self.embeddings), f'chunking_scheme': repr(self.chunking_scheme)}
        class_name = "CodeEmbeddings"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'embeddings': self.embeddings,
                'chunking_scheme': self.chunking_scheme}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class ItemStatistics(AbstractApiClass):
    """
        ItemStatistics representation.

        Args:
            client (ApiClient): An authenticated API Client instance
            missingPercent (float): percentage of missing values in data
            count (int): count of data
            median (float): median of the data
            mean (float): mean value of the data
            p10 (float): 10th percentile of the data
            p90 (float): 90th_percentile of the data
            stddev (float): standard deviation of the data
            min (int): min value in the data
            max (int): max value in the data
            lowerBound (float): lower bound threshold of the data
            upperBound (float): upper bound threshold of the data
    """

    def __init__(self, client, missingPercent=None, count=None, median=None, mean=None, p10=None, p90=None, stddev=None, min=None, max=None, lowerBound=None, upperBound=None):
        super().__init__(client, None)
        self.missing_percent = missingPercent
        self.count = count
        self.median = median
        self.mean = mean
        self.p10 = p10
        self.p90 = p90
        self.stddev = stddev
        self.min = min
        self.max = max
        self.lower_bound = lowerBound
        self.upper_bound = upperBound
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'missing_percent': repr(self.missing_percent), f'count': repr(self.count), f'median': repr(self.median), f'mean': repr(self.mean), f'p10': repr(self.p10), f'p90': repr(
            self.p90), f'stddev': repr(self.stddev), f'min': repr(self.min), f'max': repr(self.max), f'lower_bound': repr(self.lower_bound), f'upper_bound': repr(self.upper_bound)}
        class_name = "ItemStatistics"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'missing_percent': self.missing_percent, 'count': self.count, 'median': self.median, 'mean': self.mean, 'p10': self.p10,
                'p90': self.p90, 'stddev': self.stddev, 'min': self.min, 'max': self.max, 'lower_bound': self.lower_bound, 'upper_bound': self.upper_bound}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class NullViolation(AbstractApiClass):
    """
        Summary of anomalous null frequencies for a feature discovered by a model monitoring instance

        Args:
            client (ApiClient): An authenticated API Client instance
            name (str): Name of feature.
            violation (str): Description of null violation for a prediction feature.
            trainingNullFreq (float): Proportion of null entries in training feature.
            predictionNullFreq (float): Proportion of null entries in prediction feature.
    """

    def __init__(self, client, name=None, violation=None, trainingNullFreq=None, predictionNullFreq=None):
        super().__init__(client, None)
        self.name = name
        self.violation = violation
        self.training_null_freq = trainingNullFreq
        self.prediction_null_freq = predictionNullFreq
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'name': repr(self.name), f'violation': repr(self.violation), f'training_null_freq': repr(
            self.training_null_freq), f'prediction_null_freq': repr(self.prediction_null_freq)}
        class_name = "NullViolation"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'name': self.name, 'violation': self.violation, 'training_null_freq':
                self.training_null_freq, 'prediction_null_freq': self.prediction_null_freq}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class DomainPurchaseResult(AbstractApiClass):
    """
        Domain purchase result

        Args:
            client (ApiClient): An authenticated API Client instance
            domain (str): The domain name being purchased.
            billedCredits (int): The total credits used for the purchase.
    """

    def __init__(self, client, domain=None, billedCredits=None):
        super().__init__(client, None)
        self.domain = domain
        self.billed_credits = billedCredits
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'domain': repr(
            self.domain), f'billed_credits': repr(self.billed_credits)}
        class_name = "DomainPurchaseResult"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'domain': self.domain, 'billed_credits': self.billed_credits}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class ResolvedFeatureGroupTemplate(AbstractApiClass):
    """
        Final SQL from resolving a feature group template.

        Args:
            client (ApiClient): An authenticated API Client instance
            featureGroupTemplateId (str): Unique identifier for this feature group template.
            resolvedVariables (dict): Map from template variable names to parameters available during template resolution.
            resolvedSql (str): SQL resulting from resolving the SQL template by applying the resolved bindings.
            templateSql (str): SQL that can include variables to be replaced by values from the template config to resolve this template SQL into a valid SQL query for a feature group.
            sqlError (str): if invalid, the sql error message
    """

    def __init__(self, client, featureGroupTemplateId=None, resolvedVariables=None, resolvedSql=None, templateSql=None, sqlError=None):
        super().__init__(client, None)
        self.feature_group_template_id = featureGroupTemplateId
        self.resolved_variables = resolvedVariables
        self.resolved_sql = resolvedSql
        self.template_sql = templateSql
        self.sql_error = sqlError
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'feature_group_template_id': repr(self.feature_group_template_id), f'resolved_variables': repr(
            self.resolved_variables), f'resolved_sql': repr(self.resolved_sql), f'template_sql': repr(self.template_sql), f'sql_error': repr(self.sql_error)}
        class_name = "ResolvedFeatureGroupTemplate"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'feature_group_template_id': self.feature_group_template_id, 'resolved_variables': self.resolved_variables,
                'resolved_sql': self.resolved_sql, 'template_sql': self.template_sql, 'sql_error': self.sql_error}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class DaemonTaskConversation(AbstractApiClass):
    """
        Daemon Task Conversation

        Args:
            client (ApiClient): An authenticated API Client instance
            deploymentConversationId (id): The ID of the deployment conversation
            createdAt (str): The creation timestamp
            updatedAt (str): The last update timestamp
            deploymentConversationName (str): The name of the conversation
            externalApplicationId (str): The external application ID
    """

    def __init__(self, client, deploymentConversationId=None, createdAt=None, updatedAt=None, deploymentConversationName=None, externalApplicationId=None):
        super().__init__(client, None)
        self.deployment_conversation_id = deploymentConversationId
        self.created_at = createdAt
        self.updated_at = updatedAt
        self.deployment_conversation_name = deploymentConversationName
        self.external_application_id = externalApplicationId
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'deployment_conversation_id': repr(self.deployment_conversation_id), f'created_at': repr(self.created_at), f'updated_at': repr(
            self.updated_at), f'deployment_conversation_name': repr(self.deployment_conversation_name), f'external_application_id': repr(self.external_application_id)}
        class_name = "DaemonTaskConversation"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'deployment_conversation_id': self.deployment_conversation_id, 'created_at': self.created_at, 'updated_at': self.updated_at,
                'deployment_conversation_name': self.deployment_conversation_name, 'external_application_id': self.external_application_id}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class ExternalApplication(AbstractApiClass):
    """
        An external application.

        Args:
            client (ApiClient): An authenticated API Client instance
            name (str): The name of the external application.
            externalApplicationId (str): The unique identifier of the external application.
            deploymentId (str): The deployment id associated with the external application.
            description (str): The description of the external application.
            logo (str): The logo.
            theme (dict): The theme used for the External Application.
            userGroupIds (list): A list of App User Groups with access to this external application
            useCase (str): Use Case of the project of this deployment
            isAgent (bool): Whether the external application is an agent.
            status (str): The status of the deployment.
            deploymentConversationRetentionHours (int): The retention policy for the external application.
            managedUserService (str): The external service that is managing the user accounts.
            predictionOverrides (dict): The prediction overrides for the external application.
            isSystemCreated (bool): Whether the external application is system created.
            isCustomizable (bool): Whether the external application is customizable.
            isDeprecated (bool): Whether the external application is deprecated.
            isVisible (bool): Whether the external application should be shown in the dropdown.
            hasThinkingOption (bool): Whether to show the thinking option in the toolbar.
            hasFastMode (bool): Whether the external application has fast mode.
            hasEffortOption (bool): Whether the external application has an effort option.
            onlyImageGenEnabled (bool): Whether to LLM only allows image generation.
            projectId (str): The project id associated with the external application.
            isCodellmChatmodeSupported (bool): Whether the external application is codellm chatmode supported
    """

    def __init__(self, client, name=None, externalApplicationId=None, deploymentId=None, description=None, logo=None, theme=None, userGroupIds=None, useCase=None, isAgent=None, status=None, deploymentConversationRetentionHours=None, managedUserService=None, predictionOverrides=None, isSystemCreated=None, isCustomizable=None, isDeprecated=None, isVisible=None, hasThinkingOption=None, hasFastMode=None, hasEffortOption=None, onlyImageGenEnabled=None, projectId=None, isCodellmChatmodeSupported=None):
        super().__init__(client, externalApplicationId)
        self.name = name
        self.external_application_id = externalApplicationId
        self.deployment_id = deploymentId
        self.description = description
        self.logo = logo
        self.theme = theme
        self.user_group_ids = userGroupIds
        self.use_case = useCase
        self.is_agent = isAgent
        self.status = status
        self.deployment_conversation_retention_hours = deploymentConversationRetentionHours
        self.managed_user_service = managedUserService
        self.prediction_overrides = predictionOverrides
        self.is_system_created = isSystemCreated
        self.is_customizable = isCustomizable
        self.is_deprecated = isDeprecated
        self.is_visible = isVisible
        self.has_thinking_option = hasThinkingOption
        self.has_fast_mode = hasFastMode
        self.has_effort_option = hasEffortOption
        self.only_image_gen_enabled = onlyImageGenEnabled
        self.project_id = projectId
        self.is_codellm_chatmode_supported = isCodellmChatmodeSupported
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'name': repr(self.name), f'external_application_id': repr(self.external_application_id), f'deployment_id': repr(self.deployment_id), f'description': repr(self.description), f'logo': repr(self.logo), f'theme': repr(self.theme), f'user_group_ids': repr(self.user_group_ids), f'use_case': repr(self.use_case), f'is_agent': repr(self.is_agent), f'status': repr(self.status), f'deployment_conversation_retention_hours': repr(self.deployment_conversation_retention_hours), f'managed_user_service': repr(self.managed_user_service), f'prediction_overrides': repr(
            self.prediction_overrides), f'is_system_created': repr(self.is_system_created), f'is_customizable': repr(self.is_customizable), f'is_deprecated': repr(self.is_deprecated), f'is_visible': repr(self.is_visible), f'has_thinking_option': repr(self.has_thinking_option), f'has_fast_mode': repr(self.has_fast_mode), f'has_effort_option': repr(self.has_effort_option), f'only_image_gen_enabled': repr(self.only_image_gen_enabled), f'project_id': repr(self.project_id), f'is_codellm_chatmode_supported': repr(self.is_codellm_chatmode_supported)}
        class_name = "ExternalApplication"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'name': self.name, 'external_application_id': self.external_application_id, 'deployment_id': self.deployment_id, 'description': self.description, 'logo': self.logo, 'theme': self.theme, 'user_group_ids': self.user_group_ids, 'use_case': self.use_case, 'is_agent': self.is_agent, 'status': self.status, 'deployment_conversation_retention_hours': self.deployment_conversation_retention_hours, 'managed_user_service': self.managed_user_service,
                'prediction_overrides': self.prediction_overrides, 'is_system_created': self.is_system_created, 'is_customizable': self.is_customizable, 'is_deprecated': self.is_deprecated, 'is_visible': self.is_visible, 'has_thinking_option': self.has_thinking_option, 'has_fast_mode': self.has_fast_mode, 'has_effort_option': self.has_effort_option, 'only_image_gen_enabled': self.only_image_gen_enabled, 'project_id': self.project_id, 'is_codellm_chatmode_supported': self.is_codellm_chatmode_supported}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}

    def add_developers_to(self):
        """
        Adds a permission for the platform App User Group to access an External Application.

        Args:
            external_application_id (str): The ID of the External Application.
        """
        return self.client.add_developers_to_external_application(self.external_application_id)

    def remove_developers_from(self):
        """
        Removes a permission for the platform App User Group to access an External Application.

        Args:
            external_application_id (str): The ID of the External Application.
        """
        return self.client.remove_developers_from_external_application(self.external_application_id)

    def update(self, name: str = None, description: str = None, theme: dict = None, deployment_id: str = None, deployment_conversation_retention_hours: int = None, reset_retention_policy: bool = False):
        """
        Updates an External Application.

        Args:
            name (str): The name of the External Application.
            description (str): The description of the External Application. This will be shown to users when they access the External Application.
            theme (dict): The visual theme of the External Application.
            deployment_id (str): The ID of the deployment to use.
            deployment_conversation_retention_hours (int): The number of hours to retain the conversations for.
            reset_retention_policy (bool): If true, the retention policy will be removed.

        Returns:
            ExternalApplication: The updated External Application.
        """
        return self.client.update_external_application(self.external_application_id, name, description, theme, deployment_id, deployment_conversation_retention_hours, reset_retention_policy)

    def refresh(self):
        """
        Calls describe and refreshes the current object's fields

        Returns:
            ExternalApplication: The current object
        """
        self.__dict__.update(self.describe().__dict__)
        return self

    def describe(self):
        """
        Describes an External Application.

        Args:
            external_application_id (str): The ID of the External Application.

        Returns:
            ExternalApplication: The External Application.
        """
        return self.client.describe_external_application(self.external_application_id)

    def delete(self):
        """
        Deletes an External Application.

        Args:
            external_application_id (str): The ID of the External Application.
        """
        return self.client.delete_external_application(self.external_application_id)



from .external_application import ExternalApplication
from .return_class import AbstractApiClass


class BotInfo(AbstractApiClass):
    """
        Information about an external application and LLM.

        Args:
            client (ApiClient): An authenticated API Client instance
            externalApplicationId (str): The external application ID.
            deploymentId (str): The deployment ID.
            externalApplication (ExternalApplication): The described external application details.
    """

    def __init__(self, client, externalApplicationId=None, deploymentId=None, externalApplication={}):
        super().__init__(client, None)
        self.external_application_id = externalApplicationId
        self.deployment_id = deploymentId
        self.external_application = client._build_class(
            ExternalApplication, externalApplication)
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'external_application_id': repr(self.external_application_id), f'deployment_id': repr(
            self.deployment_id), f'external_application': repr(self.external_application)}
        class_name = "BotInfo"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'external_application_id': self.external_application_id, 'deployment_id': self.deployment_id,
                'external_application': self._get_attribute_as_dict(self.external_application)}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .database_connector_column import DatabaseConnectorColumn
from .return_class import AbstractApiClass


class DatabaseConnectorSchema(AbstractApiClass):
    """
        A schema description for a table from a database connector

        Args:
            client (ApiClient): An authenticated API Client instance
            tableName (str): The unique name of the table.
            columns (DatabaseConnectorColumn): List of columns in the table.
    """

    def __init__(self, client, tableName=None, columns={}):
        super().__init__(client, None)
        self.table_name = tableName
        self.columns = client._build_class(DatabaseConnectorColumn, columns)
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'table_name': repr(
            self.table_name), f'columns': repr(self.columns)}
        class_name = "DatabaseConnectorSchema"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'table_name': self.table_name,
                'columns': self._get_attribute_as_dict(self.columns)}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class ApiKey(AbstractApiClass):
    """
        An API Key to authenticate requests to the Abacus.AI API

        Args:
            client (ApiClient): An authenticated API Client instance
            apiKeyId (str): The unique ID for the API key
            apiKey (str): The unique API key scoped to a specific organization. Value will be partially obscured.
            apiKeySuffix (str): The last 4 characters of the API key.
            tag (str): A user-friendly tag for the API key.
            type (str): The type of the API key, either 'default', 'code-llm', 'terminal', or 'computer-use'.
            createdAt (str): The timestamp when the API key was created.
            expiresAt (str): The timestamp when the API key will expire.
            isExpired (bool): Whether the API key has expired.
    """

    def __init__(self, client, apiKeyId=None, apiKey=None, apiKeySuffix=None, tag=None, type=None, createdAt=None, expiresAt=None, isExpired=None):
        super().__init__(client, apiKeyId)
        self.api_key_id = apiKeyId
        self.api_key = apiKey
        self.api_key_suffix = apiKeySuffix
        self.tag = tag
        self.type = type
        self.created_at = createdAt
        self.expires_at = expiresAt
        self.is_expired = isExpired
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'api_key_id': repr(self.api_key_id), f'api_key': repr(self.api_key), f'api_key_suffix': repr(self.api_key_suffix), f'tag': repr(
            self.tag), f'type': repr(self.type), f'created_at': repr(self.created_at), f'expires_at': repr(self.expires_at), f'is_expired': repr(self.is_expired)}
        class_name = "ApiKey"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'api_key_id': self.api_key_id, 'api_key': self.api_key, 'api_key_suffix': self.api_key_suffix, 'tag': self.tag,
                'type': self.type, 'created_at': self.created_at, 'expires_at': self.expires_at, 'is_expired': self.is_expired}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}

    def delete(self):
        """
        Delete a specified API key.

        Args:
            api_key_id (str): The ID of the API key to delete.
        """
        return self.client.delete_api_key(self.api_key_id)



from .return_class import AbstractApiClass


class ModelArtifactsExport(AbstractApiClass):
    """
        A Model Artifacts Export Job

        Args:
            client (ApiClient): An authenticated API Client instance
            modelArtifactsExportId (str): Unique identifier for this export.
            modelVersion (str): Version of the model being exported.
            outputLocation (str): File Connector location the feature group is being written to.
            status (str): Current status of the export.
            createdAt (str): Timestamp at which the export was created (ISO-8601 format).
            exportCompletedAt (str): Timestamp at which the export completed (ISO-8601 format).
            error (str): If `status` is `FAILED`, this field will be populated with an error.
    """

    def __init__(self, client, modelArtifactsExportId=None, modelVersion=None, outputLocation=None, status=None, createdAt=None, exportCompletedAt=None, error=None):
        super().__init__(client, modelArtifactsExportId)
        self.model_artifacts_export_id = modelArtifactsExportId
        self.model_version = modelVersion
        self.output_location = outputLocation
        self.status = status
        self.created_at = createdAt
        self.export_completed_at = exportCompletedAt
        self.error = error
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'model_artifacts_export_id': repr(self.model_artifacts_export_id), f'model_version': repr(self.model_version), f'output_location': repr(
            self.output_location), f'status': repr(self.status), f'created_at': repr(self.created_at), f'export_completed_at': repr(self.export_completed_at), f'error': repr(self.error)}
        class_name = "ModelArtifactsExport"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'model_artifacts_export_id': self.model_artifacts_export_id, 'model_version': self.model_version, 'output_location': self.output_location,
                'status': self.status, 'created_at': self.created_at, 'export_completed_at': self.export_completed_at, 'error': self.error}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}

    def refresh(self):
        """
        Calls describe and refreshes the current object's fields

        Returns:
            ModelArtifactsExport: The current object
        """
        self.__dict__.update(self.describe().__dict__)
        return self

    def describe(self):
        """
        Get the description and status of the model artifacts export.

        Args:
            model_artifacts_export_id (str): A unique string identifier for the export.

        Returns:
            ModelArtifactsExport: Object describing the export and its status.
        """
        return self.client.describe_model_artifacts_export(self.model_artifacts_export_id)

    def wait_for_results(self, timeout=3600):
        """
        A waiting call until model artifacts export is created.

        Args:
            timeout (int): The waiting time given to the call to finish, if it doesn't finish by the allocated time, the call is said to be timed out.
        """
        return self.client._poll(self, {'PENDING', 'EXPORTING'}, timeout=timeout)

    def get_status(self):
        """
        Gets the status of the model artifacts export.

        Returns:
            str: A string describing the status of a model artifacts export (pending, complete, etc.).
        """
        return self.describe().status



from .return_class import AbstractApiClass


class VerticalFilesResult(AbstractApiClass):
    """
        Paginated result of vertical files listing.

        Args:
            client (ApiClient): An authenticated API Client instance
            files (list): List of file objects with document_upload_id, file_name, file_metadata, etc.
            maxCount (int): The total number of files.
    """

    def __init__(self, client, files=None, maxCount=None):
        super().__init__(client, None)
        self.files = files
        self.max_count = maxCount
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'files': repr(self.files),
                     f'max_count': repr(self.max_count)}
        class_name = "VerticalFilesResult"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'files': self.files, 'max_count': self.max_count}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class FeatureImportance(AbstractApiClass):
    """
        Feature importance for a specified model monitor

        Args:
            client (ApiClient): An authenticated API Client instance
            shapFeatureImportance (dict): A map of feature name to feature importance, determined by Shap values on a sample dataset.
            limeFeatureImportance (dict): A map of feature name to feature importance, determined by Lime contribution values on a sample dataset.
            permutationFeatureImportance (dict): A map of feature name to feature importance, determined by permutation importance.
            nullFeatureImportance (dict): A map of feature name to feature importance, determined by null feature importance.
            lofoFeatureImportance (dict): A map of feature name to feature importance, determined by the Leave One Feature Out method.
            ebmFeatureImportance (dict): A map of feature name to feature importance, determined by an Explainable Boosting Machine.
    """

    def __init__(self, client, shapFeatureImportance=None, limeFeatureImportance=None, permutationFeatureImportance=None, nullFeatureImportance=None, lofoFeatureImportance=None, ebmFeatureImportance=None):
        super().__init__(client, None)
        self.shap_feature_importance = shapFeatureImportance
        self.lime_feature_importance = limeFeatureImportance
        self.permutation_feature_importance = permutationFeatureImportance
        self.null_feature_importance = nullFeatureImportance
        self.lofo_feature_importance = lofoFeatureImportance
        self.ebm_feature_importance = ebmFeatureImportance
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'shap_feature_importance': repr(self.shap_feature_importance), f'lime_feature_importance': repr(self.lime_feature_importance), f'permutation_feature_importance': repr(
            self.permutation_feature_importance), f'null_feature_importance': repr(self.null_feature_importance), f'lofo_feature_importance': repr(self.lofo_feature_importance), f'ebm_feature_importance': repr(self.ebm_feature_importance)}
        class_name = "FeatureImportance"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'shap_feature_importance': self.shap_feature_importance, 'lime_feature_importance': self.lime_feature_importance, 'permutation_feature_importance': self.permutation_feature_importance,
                'null_feature_importance': self.null_feature_importance, 'lofo_feature_importance': self.lofo_feature_importance, 'ebm_feature_importance': self.ebm_feature_importance}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class ModelBlueprintStage(AbstractApiClass):
    """
        A stage in the model blueprint export process.

        Args:
            client (ApiClient): An authenticated API Client instance
            stageName (str): The name of the stage.
            displayName (str): The display name of the stage.
            description (str): The description of the stage.
            params (dict): The parameters for the stage.
            predecessors (list): A list of stages that occur directly before this stage.
    """

    def __init__(self, client, stageName=None, displayName=None, description=None, params=None, predecessors=None):
        super().__init__(client, None)
        self.stage_name = stageName
        self.display_name = displayName
        self.description = description
        self.params = params
        self.predecessors = predecessors
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'stage_name': repr(self.stage_name), f'display_name': repr(self.display_name), f'description': repr(
            self.description), f'params': repr(self.params), f'predecessors': repr(self.predecessors)}
        class_name = "ModelBlueprintStage"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'stage_name': self.stage_name, 'display_name': self.display_name,
                'description': self.description, 'params': self.params, 'predecessors': self.predecessors}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class MobileBuildInfo(AbstractApiClass):
    """
        Mobile app build information for a specific build type.

        Args:
            client (ApiClient): An authenticated API Client instance
            type (str): The type of mobile app build (APK, AAB, IOS).
            buildUrl (str): The URL to download the mobile app build.
            status (str): The status of the mobile app build.
            llmArtifactId (str): The artifact id associated with the build.
            mobileAppBuildId (str): The mobile app build id.
    """

    def __init__(self, client, type=None, buildUrl=None, status=None, llmArtifactId=None, mobileAppBuildId=None):
        super().__init__(client, None)
        self.type = type
        self.build_url = buildUrl
        self.status = status
        self.llm_artifact_id = llmArtifactId
        self.mobile_app_build_id = mobileAppBuildId
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'type': repr(self.type), f'build_url': repr(self.build_url), f'status': repr(
            self.status), f'llm_artifact_id': repr(self.llm_artifact_id), f'mobile_app_build_id': repr(self.mobile_app_build_id)}
        class_name = "MobileBuildInfo"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'type': self.type, 'build_url': self.build_url, 'status': self.status,
                'llm_artifact_id': self.llm_artifact_id, 'mobile_app_build_id': self.mobile_app_build_id}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class ChatllmComputer(AbstractApiClass):
    """
        ChatLLMComputer

        Args:
            client (ApiClient): An authenticated API Client instance
            computerId (int): The computer id.
            token (str): The token.
            vncEndpoint (str): The VNC endpoint.
    """

    def __init__(self, client, computerId=None, token=None, vncEndpoint=None):
        super().__init__(client, None)
        self.computer_id = computerId
        self.token = token
        self.vnc_endpoint = vncEndpoint
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'computer_id': repr(self.computer_id), f'token': repr(
            self.token), f'vnc_endpoint': repr(self.vnc_endpoint)}
        class_name = "ChatllmComputer"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'computer_id': self.computer_id,
                'token': self.token, 'vnc_endpoint': self.vnc_endpoint}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class EntriAuthToken(AbstractApiClass):
    """
        Entri Auth Token

        Args:
            client (ApiClient): An authenticated API Client instance
            token (str): The authentication token for Entri
            applicationId (str): application Id from Entri dashboard
            ttl (int): The duration in milliseconds for which the token is valid
    """

    def __init__(self, client, token=None, applicationId=None, ttl=None):
        super().__init__(client, None)
        self.token = token
        self.application_id = applicationId
        self.ttl = ttl
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'token': repr(self.token), f'application_id': repr(
            self.application_id), f'ttl': repr(self.ttl)}
        class_name = "EntriAuthToken"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'token': self.token,
                'application_id': self.application_id, 'ttl': self.ttl}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .daemon_task_instance import DaemonTaskInstance
from .hosted_database import HostedDatabase
from .return_class import AbstractApiClass


class ChatllmTask(AbstractApiClass):
    """
        A chatllm task

        Args:
            client (ApiClient): An authenticated API Client instance
            chatllmTaskId (str): The id of the chatllm task.
            daemonTaskId (str): The id of the daemon task.
            taskType (str): The type of task ('chatllm' or 'daemon').
            name (str): The name of the chatllm task.
            instructions (str): The instructions of the chatllm task.
            description (str): The description of the chatllm task.
            lifecycle (str): The lifecycle of the chatllm task.
            scheduleInfo (dict): The schedule info of the chatllm task.
            externalApplicationId (str): The external application id associated with the chatllm task.
            deploymentConversationId (str): The deployment conversation id associated with the chatllm task.
            sourceDeploymentConversationId (str): The source deployment conversation id associated with the chatllm task.
            enableEmailAlerts (bool): Whether email alerts are enabled for the chatllm task.
            email (str): The email to send alerts to.
            numUnreadTaskInstances (int): The number of unread task instances for the chatllm task.
            computePointsUsed (int): The compute points used for the chatllm task.
            displayMarkdown (str): The display markdown for the chatllm task.
            requiresNewConversation (bool): Whether a new conversation is required for the chatllm task.
            executionMode (str): The execution mode of the chatllm task.
            taskDefinition (dict): The task definition (for web_service_trigger tasks).
            webAppHostname (str): The hostname of the web app associated with the daemon task.
            triggerType (str): The trigger type of the daemon task (scheduled or event_based).
            webhookUrl (str): The webhook URL for event-based daemon tasks.
            pushNotificationsEnabled (bool): Whether push notifications are enabled for the task.
            hostedDatabase (HostedDatabase): The hosted database for the daemon task.
            latestDaemonTaskInstance (DaemonTaskInstance): The latest task instance for daemon tasks.
    """

    def __init__(self, client, chatllmTaskId=None, daemonTaskId=None, taskType=None, name=None, instructions=None, description=None, lifecycle=None, scheduleInfo=None, externalApplicationId=None, deploymentConversationId=None, sourceDeploymentConversationId=None, enableEmailAlerts=None, email=None, numUnreadTaskInstances=None, computePointsUsed=None, displayMarkdown=None, requiresNewConversation=None, executionMode=None, taskDefinition=None, webAppHostname=None, triggerType=None, webhookUrl=None, pushNotificationsEnabled=None, hostedDatabase={}, latestDaemonTaskInstance={}):
        super().__init__(client, chatllmTaskId)
        self.chatllm_task_id = chatllmTaskId
        self.daemon_task_id = daemonTaskId
        self.task_type = taskType
        self.name = name
        self.instructions = instructions
        self.description = description
        self.lifecycle = lifecycle
        self.schedule_info = scheduleInfo
        self.external_application_id = externalApplicationId
        self.deployment_conversation_id = deploymentConversationId
        self.source_deployment_conversation_id = sourceDeploymentConversationId
        self.enable_email_alerts = enableEmailAlerts
        self.email = email
        self.num_unread_task_instances = numUnreadTaskInstances
        self.compute_points_used = computePointsUsed
        self.display_markdown = displayMarkdown
        self.requires_new_conversation = requiresNewConversation
        self.execution_mode = executionMode
        self.task_definition = taskDefinition
        self.web_app_hostname = webAppHostname
        self.trigger_type = triggerType
        self.webhook_url = webhookUrl
        self.push_notifications_enabled = pushNotificationsEnabled
        self.hosted_database = client._build_class(
            HostedDatabase, hostedDatabase)
        self.latest_daemon_task_instance = client._build_class(
            DaemonTaskInstance, latestDaemonTaskInstance)
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'chatllm_task_id': repr(self.chatllm_task_id), f'daemon_task_id': repr(self.daemon_task_id), f'task_type': repr(self.task_type), f'name': repr(self.name), f'instructions': repr(self.instructions), f'description': repr(self.description), f'lifecycle': repr(self.lifecycle), f'schedule_info': repr(self.schedule_info), f'external_application_id': repr(self.external_application_id), f'deployment_conversation_id': repr(self.deployment_conversation_id), f'source_deployment_conversation_id': repr(self.source_deployment_conversation_id), f'enable_email_alerts': repr(self.enable_email_alerts), f'email': repr(self.email), f'num_unread_task_instances': repr(
            self.num_unread_task_instances), f'compute_points_used': repr(self.compute_points_used), f'display_markdown': repr(self.display_markdown), f'requires_new_conversation': repr(self.requires_new_conversation), f'execution_mode': repr(self.execution_mode), f'task_definition': repr(self.task_definition), f'web_app_hostname': repr(self.web_app_hostname), f'trigger_type': repr(self.trigger_type), f'webhook_url': repr(self.webhook_url), f'push_notifications_enabled': repr(self.push_notifications_enabled), f'hosted_database': repr(self.hosted_database), f'latest_daemon_task_instance': repr(self.latest_daemon_task_instance)}
        class_name = "ChatllmTask"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'chatllm_task_id': self.chatllm_task_id, 'daemon_task_id': self.daemon_task_id, 'task_type': self.task_type, 'name': self.name, 'instructions': self.instructions, 'description': self.description, 'lifecycle': self.lifecycle, 'schedule_info': self.schedule_info, 'external_application_id': self.external_application_id, 'deployment_conversation_id': self.deployment_conversation_id, 'source_deployment_conversation_id': self.source_deployment_conversation_id, 'enable_email_alerts': self.enable_email_alerts, 'email': self.email, 'num_unread_task_instances': self.num_unread_task_instances,
                'compute_points_used': self.compute_points_used, 'display_markdown': self.display_markdown, 'requires_new_conversation': self.requires_new_conversation, 'execution_mode': self.execution_mode, 'task_definition': self.task_definition, 'web_app_hostname': self.web_app_hostname, 'trigger_type': self.trigger_type, 'webhook_url': self.webhook_url, 'push_notifications_enabled': self.push_notifications_enabled, 'hosted_database': self._get_attribute_as_dict(self.hosted_database), 'latest_daemon_task_instance': self._get_attribute_as_dict(self.latest_daemon_task_instance)}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class DomainSearchResult(AbstractApiClass):
    """
        Domain search result

        Args:
            client (ApiClient): An authenticated API Client instance
            domain (str): name of the domain in a search result
            registerCredits (int): number of credits required for registration
            registerYears (int): number of years the domain will get registered for
            renewalCredits (int): number of credits required for renewal
            renewalYears (int): number of years the domain will be renewed for
    """

    def __init__(self, client, domain=None, registerCredits=None, registerYears=None, renewalCredits=None, renewalYears=None):
        super().__init__(client, None)
        self.domain = domain
        self.register_credits = registerCredits
        self.register_years = registerYears
        self.renewal_credits = renewalCredits
        self.renewal_years = renewalYears
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'domain': repr(self.domain), f'register_credits': repr(self.register_credits), f'register_years': repr(
            self.register_years), f'renewal_credits': repr(self.renewal_credits), f'renewal_years': repr(self.renewal_years)}
        class_name = "DomainSearchResult"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'domain': self.domain, 'register_credits': self.register_credits, 'register_years': self.register_years,
                'renewal_credits': self.renewal_credits, 'renewal_years': self.renewal_years}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .forecasting_analysis_graph_data import ForecastingAnalysisGraphData
from .return_class import AbstractApiClass


class EdaForecastingAnalysis(AbstractApiClass):
    """
        Eda Forecasting Analysis of the latest version of the data.

        Args:
            client (ApiClient): An authenticated API Client instance
            primaryKeys (list): Name of the primary keys in the data
            forecastingTargetFeature (str): Feature in the data that represents the target.
            timestampFeature (str): Feature in the data that represents the timestamp column.
            forecastFrequency (str): Frequency of data, could be hourly, daily, weekly, monthly, quarterly or yearly.
            salesAcrossTime (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across time
            cummulativeContribution (ForecastingAnalysisGraphData): Data showing what percent of items contribute to what amount of sales.
            missingValueDistribution (ForecastingAnalysisGraphData): Data showing missing or null value distribution
            historyLength (ForecastingAnalysisGraphData): Data showing length of history distribution
            numRowsHistogram (ForecastingAnalysisGraphData): Data showing number of rows for an item distribution
            productMaturity (ForecastingAnalysisGraphData): Data showing length of how long a product has been alive with average, p10, p90 and median
            seasonalityYear (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across grouped years
            seasonalityMonth (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across grouped months
            seasonalityWeekOfYear (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across week of year seasonality
            seasonalityDayOfYear (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across day of year seasonality
            seasonalityDayOfMonth (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across day of month seasonality
            seasonalityDayOfWeek (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across day of week seasonality
            seasonalityQuarter (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across grouped quarters
            seasonalityHour (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across grouped hours
            seasonalityMinute (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across grouped minutes
            seasonalitySecond (ForecastingAnalysisGraphData): Data showing average, p10, p90, median sales across grouped seconds
            autocorrelation (ForecastingAnalysisGraphData): Data showing the correlation of the forecasting target and its lagged values at different time lags.
            partialAutocorrelation (ForecastingAnalysisGraphData): Data showing the correlation of the forecasting target and its lagged values, controlling for the effects of intervening lags.
    """

    def __init__(self, client, primaryKeys=None, forecastingTargetFeature=None, timestampFeature=None, forecastFrequency=None, salesAcrossTime={}, cummulativeContribution={}, missingValueDistribution={}, historyLength={}, numRowsHistogram={}, productMaturity={}, seasonalityYear={}, seasonalityMonth={}, seasonalityWeekOfYear={}, seasonalityDayOfYear={}, seasonalityDayOfMonth={}, seasonalityDayOfWeek={}, seasonalityQuarter={}, seasonalityHour={}, seasonalityMinute={}, seasonalitySecond={}, autocorrelation={}, partialAutocorrelation={}):
        super().__init__(client, None)
        self.primary_keys = primaryKeys
        self.forecasting_target_feature = forecastingTargetFeature
        self.timestamp_feature = timestampFeature
        self.forecast_frequency = forecastFrequency
        self.sales_across_time = client._build_class(
            ForecastingAnalysisGraphData, salesAcrossTime)
        self.cummulative_contribution = client._build_class(
            ForecastingAnalysisGraphData, cummulativeContribution)
        self.missing_value_distribution = client._build_class(
            ForecastingAnalysisGraphData, missingValueDistribution)
        self.history_length = client._build_class(
            ForecastingAnalysisGraphData, historyLength)
        self.num_rows_histogram = client._build_class(
            ForecastingAnalysisGraphData, numRowsHistogram)
        self.product_maturity = client._build_class(
            ForecastingAnalysisGraphData, productMaturity)
        self.seasonality_year = client._build_class(
            ForecastingAnalysisGraphData, seasonalityYear)
        self.seasonality_month = client._build_class(
            ForecastingAnalysisGraphData, seasonalityMonth)
        self.seasonality_week_of_year = client._build_class(
            ForecastingAnalysisGraphData, seasonalityWeekOfYear)
        self.seasonality_day_of_year = client._build_class(
            ForecastingAnalysisGraphData, seasonalityDayOfYear)
        self.seasonality_day_of_month = client._build_class(
            ForecastingAnalysisGraphData, seasonalityDayOfMonth)
        self.seasonality_day_of_week = client._build_class(
            ForecastingAnalysisGraphData, seasonalityDayOfWeek)
        self.seasonality_quarter = client._build_class(
            ForecastingAnalysisGraphData, seasonalityQuarter)
        self.seasonality_hour = client._build_class(
            ForecastingAnalysisGraphData, seasonalityHour)
        self.seasonality_minute = client._build_class(
            ForecastingAnalysisGraphData, seasonalityMinute)
        self.seasonality_second = client._build_class(
            ForecastingAnalysisGraphData, seasonalitySecond)
        self.autocorrelation = client._build_class(
            ForecastingAnalysisGraphData, autocorrelation)
        self.partial_autocorrelation = client._build_class(
            ForecastingAnalysisGraphData, partialAutocorrelation)
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'primary_keys': repr(self.primary_keys), f'forecasting_target_feature': repr(self.forecasting_target_feature), f'timestamp_feature': repr(self.timestamp_feature), f'forecast_frequency': repr(self.forecast_frequency), f'sales_across_time': repr(self.sales_across_time), f'cummulative_contribution': repr(self.cummulative_contribution), f'missing_value_distribution': repr(self.missing_value_distribution), f'history_length': repr(self.history_length), f'num_rows_histogram': repr(self.num_rows_histogram), f'product_maturity': repr(self.product_maturity), f'seasonality_year': repr(self.seasonality_year), f'seasonality_month': repr(
            self.seasonality_month), f'seasonality_week_of_year': repr(self.seasonality_week_of_year), f'seasonality_day_of_year': repr(self.seasonality_day_of_year), f'seasonality_day_of_month': repr(self.seasonality_day_of_month), f'seasonality_day_of_week': repr(self.seasonality_day_of_week), f'seasonality_quarter': repr(self.seasonality_quarter), f'seasonality_hour': repr(self.seasonality_hour), f'seasonality_minute': repr(self.seasonality_minute), f'seasonality_second': repr(self.seasonality_second), f'autocorrelation': repr(self.autocorrelation), f'partial_autocorrelation': repr(self.partial_autocorrelation)}
        class_name = "EdaForecastingAnalysis"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'primary_keys': self.primary_keys, 'forecasting_target_feature': self.forecasting_target_feature, 'timestamp_feature': self.timestamp_feature, 'forecast_frequency': self.forecast_frequency, 'sales_across_time': self._get_attribute_as_dict(self.sales_across_time), 'cummulative_contribution': self._get_attribute_as_dict(self.cummulative_contribution), 'missing_value_distribution': self._get_attribute_as_dict(self.missing_value_distribution), 'history_length': self._get_attribute_as_dict(self.history_length), 'num_rows_histogram': self._get_attribute_as_dict(self.num_rows_histogram), 'product_maturity': self._get_attribute_as_dict(self.product_maturity), 'seasonality_year': self._get_attribute_as_dict(self.seasonality_year), 'seasonality_month': self._get_attribute_as_dict(
            self.seasonality_month), 'seasonality_week_of_year': self._get_attribute_as_dict(self.seasonality_week_of_year), 'seasonality_day_of_year': self._get_attribute_as_dict(self.seasonality_day_of_year), 'seasonality_day_of_month': self._get_attribute_as_dict(self.seasonality_day_of_month), 'seasonality_day_of_week': self._get_attribute_as_dict(self.seasonality_day_of_week), 'seasonality_quarter': self._get_attribute_as_dict(self.seasonality_quarter), 'seasonality_hour': self._get_attribute_as_dict(self.seasonality_hour), 'seasonality_minute': self._get_attribute_as_dict(self.seasonality_minute), 'seasonality_second': self._get_attribute_as_dict(self.seasonality_second), 'autocorrelation': self._get_attribute_as_dict(self.autocorrelation), 'partial_autocorrelation': self._get_attribute_as_dict(self.partial_autocorrelation)}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .llm_code_block import LlmCodeBlock
from .return_class import AbstractApiClass


class LlmResponse(AbstractApiClass):
    """
        The response returned by LLM

        Args:
            client (ApiClient): An authenticated API Client instance
            content (str): Full response from LLM.
            tokens (int): The number of tokens in the response.
            stopReason (str): The reason due to which the response generation stopped.
            llmName (str): The name of the LLM model used to generate the response.
            inputTokens (int): The number of input tokens used in the LLM call.
            outputTokens (int): The number of output tokens generated in the LLM response.
            totalTokens (int): The total number of tokens (input + output) used in the LLM interaction.
            codeBlocks (LlmCodeBlock): A list of parsed code blocks from raw LLM Response
    """

    def __init__(self, client, content=None, tokens=None, stopReason=None, llmName=None, inputTokens=None, outputTokens=None, totalTokens=None, codeBlocks={}):
        super().__init__(client, None)
        self.content = content
        self.tokens = tokens
        self.stop_reason = stopReason
        self.llm_name = llmName
        self.input_tokens = inputTokens
        self.output_tokens = outputTokens
        self.total_tokens = totalTokens
        self.code_blocks = client._build_class(LlmCodeBlock, codeBlocks)
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'content': repr(self.content), f'tokens': repr(self.tokens), f'stop_reason': repr(self.stop_reason), f'llm_name': repr(self.llm_name), f'input_tokens': repr(
            self.input_tokens), f'output_tokens': repr(self.output_tokens), f'total_tokens': repr(self.total_tokens), f'code_blocks': repr(self.code_blocks)}
        class_name = "LlmResponse"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'content': self.content, 'tokens': self.tokens, 'stop_reason': self.stop_reason, 'llm_name': self.llm_name, 'input_tokens': self.input_tokens,
                'output_tokens': self.output_tokens, 'total_tokens': self.total_tokens, 'code_blocks': self._get_attribute_as_dict(self.code_blocks)}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class MobileAppBuildStatus(AbstractApiClass):
    """
        Status and details of a mobile app build.

        Args:
            client (ApiClient): An authenticated API Client instance
            status (str): Current build status ('PENDING', 'SUCCESS', 'FAILED', 'CANCELLED')
            buildUrl (str): URL to download the built artifact when SUCCESS
            mobileAppBuildId (str): build identifier
            hostname (str): The hostname associated with the build
            requiredInput (str): The required input for the build
            error (str): The error message for the build
    """

    def __init__(self, client, status=None, buildUrl=None, mobileAppBuildId=None, hostname=None, requiredInput=None, error=None):
        super().__init__(client, None)
        self.status = status
        self.build_url = buildUrl
        self.mobile_app_build_id = mobileAppBuildId
        self.hostname = hostname
        self.required_input = requiredInput
        self.error = error
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'status': repr(self.status), f'build_url': repr(self.build_url), f'mobile_app_build_id': repr(
            self.mobile_app_build_id), f'hostname': repr(self.hostname), f'required_input': repr(self.required_input), f'error': repr(self.error)}
        class_name = "MobileAppBuildStatus"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'status': self.status, 'build_url': self.build_url, 'mobile_app_build_id': self.mobile_app_build_id,
                'hostname': self.hostname, 'required_input': self.required_input, 'error': self.error}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



from .return_class import AbstractApiClass


class WebAppDeployment(AbstractApiClass):
    """
        Web app deployment.

        Args:
            client (ApiClient): An authenticated API Client instance
            webAppDeploymentId (id): The ID of the web app deployment.
            hostname (str): The hostname of the web app deployment.
            accessLevel (str): The access level of the web app deployment.
            llmArtifactId (id): The ID of the LLM artifact.
            artifactsPath (str): The path to the artifacts of the web app deployment.
            applicationType (str): The type of application.
            memoryGb (float): The memory in GB of the web app deployment.
    """

    def __init__(self, client, webAppDeploymentId=None, hostname=None, accessLevel=None, llmArtifactId=None, artifactsPath=None, applicationType=None, memoryGb=None):
        super().__init__(client, webAppDeploymentId)
        self.web_app_deployment_id = webAppDeploymentId
        self.hostname = hostname
        self.access_level = accessLevel
        self.llm_artifact_id = llmArtifactId
        self.artifacts_path = artifactsPath
        self.application_type = applicationType
        self.memory_gb = memoryGb
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'web_app_deployment_id': repr(self.web_app_deployment_id), f'hostname': repr(self.hostname), f'access_level': repr(self.access_level), f'llm_artifact_id': repr(
            self.llm_artifact_id), f'artifacts_path': repr(self.artifacts_path), f'application_type': repr(self.application_type), f'memory_gb': repr(self.memory_gb)}
        class_name = "WebAppDeployment"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'web_app_deployment_id': self.web_app_deployment_id, 'hostname': self.hostname, 'access_level': self.access_level,
                'llm_artifact_id': self.llm_artifact_id, 'artifacts_path': self.artifacts_path, 'application_type': self.application_type, 'memory_gb': self.memory_gb}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}



import importlib
import json
import os
import re
import string
import uuid
from enum import Enum
from itertools import groupby
from typing import IO, Callable, List

import pandas as pd


INVALID_PANDAS_COLUMN_NAME_CHARACTERS = '[^A-Za-z0-9_]'


def clean_column_name(column):
    cleaned_col = re.sub(
        INVALID_PANDAS_COLUMN_NAME_CHARACTERS, '_', column).lstrip('_')
    if cleaned_col and cleaned_col[0] not in string.ascii_letters:
        cleaned_col = 'Column_' + cleaned_col
    return cleaned_col


def avro_to_pandas_dtype(avro_type):
    avro_pandas_dtypes = {
        'long': 'Int64',
        'int': 'Int32',
        'float': 'float32',
        'double': 'float64',
        'string': 'object',
        'boolean': 'bool',
        'bytes': 'object',
        'null': 'object',
        'date': 'object',
        'datetime': None,
    }

    if isinstance(avro_type, dict):
        if avro_type.get('logicalType') in ['timestamp-micros']:
            avro_type = 'datetime'
        elif avro_type.get('logicalType') in ['date']:
            avro_type = 'date'
        else:
            avro_type = avro_type['type']

    return avro_pandas_dtypes.get(avro_type, 'object')


def _get_spark_incompatible_columns(df):
    # Spark-compatible pandas dtypes
    spark_compatible_pd_dtypes = {
        'int8', 'int16', 'int32', 'int64',
        'float32', 'float64',
        'bool',           # Standard boolean type
        'boolean',        # Nullable BooleanDtype
        'object',         # Assuming they contain strings
        'string',         # StringDtype introduced in pandas 1.0
        'datetime64[ns]',
        'timedelta64[ns]'
    }

    incompatible_columns = []

    for col in df.columns:
        dtype = df[col].dtype
        dtype_str = str(dtype)
        if pd.api.types.is_extension_array_dtype(dtype):
            dtype_name = dtype.name.lower()
            if dtype_name not in spark_compatible_pd_dtypes:
                incompatible_columns.append((col, dtype_name))
        elif dtype_str not in spark_compatible_pd_dtypes:
            incompatible_columns.append((col, dtype_str))

    return incompatible_columns, spark_compatible_pd_dtypes


def get_non_nullable_type(types):
    non_nullable_types = [
        avro_type for avro_type in types if avro_type != 'null']
    return non_nullable_types[0] if non_nullable_types else None


class StreamingHandler(str):
    def __new__(cls, value, context=None, section_key=None, data_type='text', is_transient=False):
        if context:
            cls.process_streaming_data(
                value, context, section_key, data_type, is_transient)
        return str.__new__(cls, value)

    @classmethod
    def process_streaming_data(cls, value, context, section_key, data_type, is_transient):
        if hasattr(context, 'streamed_section_response') and hasattr(context, 'streamed_response') and not is_transient:
            if data_type == 'text':
                if section_key:
                    entry_exists = False
                    for i, item in enumerate(context.streamed_section_response):
                        if item['id'] == section_key:
                            if ((isinstance(context.streamed_section_response[i]['contents'], str) and isinstance(value, str)) or
                               (isinstance(context.streamed_section_response[i]['contents'], list) and isinstance(value, list))):
                                context.streamed_section_response[i]['contents'] += value
                            else:
                                context.streamed_section_response[i]['contents'] = value
                            entry_exists = True
                            break
                    if not entry_exists:
                        context.streamed_section_response.append(
                            {'id': section_key, 'type': data_type, 'mime_type': 'text/plain', 'contents': value, 'message_id': str(uuid.uuid4())})
                else:
                    context.streamed_response.append(str(value))
            elif data_type == 'segment':
                context.streamed_section_response.append(value)


def get_object_from_context(client, context, variable_name, return_type):
    raw_value = getattr(context, variable_name, None)
    if raw_value is None:
        return None

    from typing import _GenericAlias

    is_container_type = isinstance(return_type, _GenericAlias)
    if not is_container_type and isinstance(raw_value, return_type):
        return raw_value

    typed_value = raw_value

    #
    # Attempt to cast json strings and dicts into api class objects
    #
    try:
        from .return_class import AbstractApiClass

        list_container = return_type is list
        dict_container = return_type is dict
        base_type = return_type
        if is_container_type:
            dict_container = return_type.__origin__ is dict
            list_container = return_type.__origin__ is list
            if hasattr(return_type.__args__[-1], '__bases__'):
                base_type = return_type.__args__[-1]

        is_api_class = issubclass(base_type, AbstractApiClass)
        if isinstance(raw_value, str) and (is_api_class or list_container or dict_container):
            typed_value = json.loads(raw_value)

        if is_api_class:
            if list_container and hasattr(typed_value, '__iter__') and isinstance(next(iter(typed_value)), dict):
                typed_value = [base_type(client, **o) for o in typed_value]
            elif dict_container and isinstance(typed_value, dict) and isinstance(next(iter(typed_value.values())), dict):
                typed_value = {k: base_type(client, **v)
                               for k, v in typed_value.items()}
            elif not list_container and not dict_container and isinstance(typed_value, dict):
                typed_value = base_type(client, **typed_value)

    except Exception:
        pass

    return typed_value


def load_as_pandas_from_avro_fd(fd: IO):
    import fastavro
    import pandas as pd

    reader = fastavro.reader(fd)
    schema = reader.writer_schema
    col_dtypes = {}
    for field in schema['fields']:
        field_name = field['name']
        field_type = field['type']
        if isinstance(field_type, list):
            field_type = get_non_nullable_type(field_type)
        pandas_dtype = avro_to_pandas_dtype(field_type)
        col_dtypes[field_name] = pandas_dtype
    df_part = pd.DataFrame.from_records(
        [r for r in reader], columns=col_dtypes.keys())

    for col in df_part.columns:
        if col_dtypes[col] is None:
            df_part[col] = pd.to_datetime(
                df_part[col], errors='coerce')

        if pd.api.types.is_datetime64_any_dtype(df_part[col]):
            df_part[col] = df_part[col].dt.tz_localize(
                None)
        elif col_dtypes[col] is not None and str(df_part[col].dtype).lower() != str(col_dtypes[col]).lower():
            df_part[col] = df_part[col].astype(
                col_dtypes[col])
    return df_part


def load_as_pandas_from_avro_files(files: List[str], download_method: Callable, max_workers: int = 10):
    import tempfile
    from concurrent.futures import ThreadPoolExecutor

    import pandas as pd

    data_df = pd.DataFrame()
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        with tempfile.TemporaryDirectory() as tmp_dir:
            df_parts = []
            file_futures = [executor.submit(
                download_method, file_part, tmp_dir, i) for i, file_part in enumerate(files)]
            for future in file_futures:
                part_path = future.result()
                with open(part_path, 'rb') as part_data:
                    df_part = load_as_pandas_from_avro_fd(part_data)
                    df_parts.append(df_part)
        data_df = pd.concat(df_parts, ignore_index=True)

    return data_df


def validate_workflow_node_inputs(nodes_info, agent_workflow_node_id, keyword_arguments: dict, sample_user_inputs: dict, filtered_workflow_vars: dict):
    from .api_class import WorkflowNodeInputType
    input_mappings = nodes_info[agent_workflow_node_id].get(
        'input_mappings', {})
    input_schema = nodes_info[agent_workflow_node_id].get('input_schema', {})
    if input_schema.get('runtime_schema', False):
        keyword_arguments = {input_schema.get(
            'schema_prop'): keyword_arguments}
    for input_mapping in input_mappings:
        input_name = input_mapping['name']
        variable_type = input_mapping['variable_type']
        is_required = input_mapping.get('is_required', True)
        variable_source = input_mapping['variable_source']
        if variable_type == 'WORKFLOW_VARIABLE':
            if variable_source not in filtered_workflow_vars:
                raise ValueError(
                    f'The stage corresponding to "{agent_workflow_node_id}" requires variables from {variable_source} stage which are not there.')
            if input_name not in filtered_workflow_vars[variable_source] and is_required:
                raise ValueError(
                    f'Missing required input "{input_name}" in workflow vars for workflow node "{agent_workflow_node_id}".')
            else:
                keyword_arguments[input_name] = filtered_workflow_vars[variable_source][input_name]
        elif variable_type == WorkflowNodeInputType.USER_INPUT:
            if sample_user_inputs and input_name in sample_user_inputs:
                keyword_arguments[input_name] = sample_user_inputs[input_name]
            elif variable_source in filtered_workflow_vars and input_name in filtered_workflow_vars[variable_source]:
                keyword_arguments[input_name] = filtered_workflow_vars[variable_source][input_name]
            else:
                if is_required:
                    raise ValueError(
                        f'User input for "{input_name}" is required for the "{agent_workflow_node_id}" node.')
                else:
                    keyword_arguments[input_name] = None


def run(nodes: List[dict], primary_start_node: str, graph_info: dict, sample_user_inputs: dict = None, agent_workflow_node_id: str = None, workflow_vars: dict = {}, topological_dfs_stack: List = []):
    from .api_class import WorkflowNodeInputType
    source_code = graph_info['source_code']
    exec(source_code, globals())

    nodes_info: dict = {node['name']: node for node in nodes}
    traversal_orders = graph_info['traversal_orders']
    nodes_ancestors = graph_info['nodes_ancestors']
    nodes_inedges = graph_info['nodes_inedges']
    primary_start_node = primary_start_node or graph_info['default_root_node']

    primary_traversal_order = traversal_orders[primary_start_node]
    run_info = {}
    workflow_vars = workflow_vars.copy()
    next_agent_workflow_node_id = None

    if agent_workflow_node_id:
        next_agent_workflow_node_id = agent_workflow_node_id
        if next_agent_workflow_node_id not in traversal_orders.keys():
            if next_agent_workflow_node_id not in nodes_info:
                raise ValueError(
                    f'The provided workflow node id "{next_agent_workflow_node_id}" is not part of the workflow. Please provide a valid node id.')
            else:
                topological_dfs_stack.append(next_agent_workflow_node_id)
        else:
            topological_dfs_stack = [next_agent_workflow_node_id]
    else:
        next_agent_workflow_node_id = primary_start_node
        topological_dfs_stack = [primary_start_node]

    flow_traversal_order = primary_traversal_order
    for root, traversal_order in traversal_orders.items():
        if next_agent_workflow_node_id in traversal_order:
            flow_traversal_order = traversal_order
            break

    run_history = []
    workflow_node_outputs = {}
    while (True):
        agent_workflow_node_id = next_agent_workflow_node_id
        node_ancestors = nodes_ancestors[agent_workflow_node_id]

        # To ensure the node takes inputs only from it's ancestors.
        # workflow_vars must always contain an entry for ancestor, the error is somewhere else if this ever errors out.
        filtered_workflow_vars = {}
        for ancestor in node_ancestors:
            if ancestor not in workflow_vars:
                raise ValueError(
                    f'Ancestor "{ancestor}" of node "{agent_workflow_node_id}" is not executed yet. Please make sure the ancestor nodes are executed before the current node.')
            else:
                filtered_workflow_vars[ancestor] = workflow_vars[ancestor]

        arguments = []
        keyword_arguments = {}
        validate_workflow_node_inputs(nodes_info, agent_workflow_node_id,
                                      keyword_arguments, sample_user_inputs, filtered_workflow_vars)

        try:
            func = eval(nodes_info[agent_workflow_node_id]['function_name'])
            node_response = func(*arguments, **keyword_arguments)
            workflow_node_outputs[agent_workflow_node_id] = node_response.to_dict(
            )
            node_workflow_vars = process_node_response(node_response)
        except Exception as error:
            raise ValueError(
                f'Error in running workflow node {agent_workflow_node_id}: {error}')

        workflow_vars[agent_workflow_node_id] = node_workflow_vars
        next_agent_workflow_node_id = None
        needs_user_input = False

        potential_next_index = flow_traversal_order.index(
            topological_dfs_stack[-1]) + 1
        potential_next_agent_workflow_node_id = None
        while (potential_next_index < len(flow_traversal_order)):
            potential_next_agent_workflow_node_id = flow_traversal_order[potential_next_index]
            incoming_edges = nodes_inedges[potential_next_agent_workflow_node_id]
            valid_next_node = True
            for source, _, details in incoming_edges:
                if source not in topological_dfs_stack:
                    valid_next_node = False
                    potential_next_index += 1
                    break
                else:
                    edge_evaluate_result = evaluate_edge_condition(
                        source, potential_next_agent_workflow_node_id, details, workflow_vars)
                    if not edge_evaluate_result:
                        valid_next_node = False
                        potential_next_index += 1
                        break
            if valid_next_node:
                next_agent_workflow_node_id = potential_next_agent_workflow_node_id
                break

        if next_agent_workflow_node_id:
            next_node_input_mappings = nodes_info[next_agent_workflow_node_id].get(
                'input_mappings', [])
            needs_user_input = any([input_mapping['variable_type'] ==
                                   WorkflowNodeInputType.USER_INPUT for input_mapping in next_node_input_mappings])

        if needs_user_input:
            run_history.append(
                f'Workflow node {agent_workflow_node_id} completed with next node {next_agent_workflow_node_id} and needs user_inputs')
        else:
            run_history.append(
                f'Workflow node {agent_workflow_node_id} completed with next node {next_agent_workflow_node_id}')
            topological_dfs_stack.append(next_agent_workflow_node_id)

        if next_agent_workflow_node_id is None or needs_user_input:
            break

    run_info['workflow_node_outputs'] = workflow_node_outputs
    run_info['run_history'] = run_history

    workflow_info = {}
    workflow_info['workflow_vars'] = workflow_vars
    workflow_info['topological_dfs_stack'] = topological_dfs_stack
    workflow_info['run_info'] = run_info

    return workflow_info


def evaluate_edge_condition(source, target, details, workflow_vars):
    try:
        condition = details.get('EXECUTION_CONDITION')
        if condition:
            result = execute_python_source(
                condition, workflow_vars.get(source, {}))
            return result
        return True
    except Exception as e:
        raise ValueError(
            f"Error evaluating edge '{source}'-->'{target}': {str(e)}")


def execute_python_source(python_expression, variables):
    try:
        # Evaluate the expression using the variables dictionary
        result = eval(python_expression, {}, variables)
        return result
    except Exception as e:
        # Handle any exceptions that may occur during evaluation
        raise ValueError(f'Error evaluating expression: {e}')


def process_node_response(node_response):
    output_vars = {}
    for variable in node_response.section_data_list:
        for key, value in variable.items():
            output_vars[key] = value
    return output_vars


class StreamType(Enum):
    MESSAGE = 'message'
    SECTION_OUTPUT = 'section_output'
    SEGMENT = 'segment'


class DocstoreUtils:
    """Utility class for loading docstore data.
    Needs to be updated if docstore formats change."""

    DOC_ID = 'doc_id'
    PREDICTION_PREFIX = 'prediction'
    FIRST_PAGE = 'first_page'
    LAST_PAGE = 'last_page'
    PAGE_TEXT = 'page_text'
    PAGES = 'pages'
    CONTENT = 'content'
    TOKENS = 'tokens'
    PAGES_ZIP_METADATA = 'pages_zip_metadata'
    PAGE_DATA = 'page_data'
    HEIGHT = 'height'
    WIDTH = 'width'
    METADATA = 'metadata'
    PAGE = 'page'
    BLOCK = 'block'
    LINE = 'line'
    EXTRACTED_TEXT = 'extracted_text'
    EMBEDDED_TEXT = 'embedded_text'
    PAGE_MARKDOWN = 'page_markdown'
    PAGE_LLM_OCR = 'page_llm_ocr'
    PAGE_TABLE_TEXT = 'page_table_text'
    MARKDOWN_FEATURES = 'markdown_features'
    MULTI_MODE_OCR_TEXT = 'multi_mode_ocr_text'
    DOCUMENT_PROCESSING_CONFIG = 'document_processing_config'
    DOCUMENT_PROCESSING_VERSION = 'document_processing_version'

    @staticmethod
    def get_archive_id(doc_id: str):
        parts = doc_id.split('-')
        if len(parts) < 3:
            raise ValueError(f'Unsupported doc_id: {doc_id}')

        if parts[0] == DocstoreUtils.PREDICTION_PREFIX:
            raise ValueError(f'Unsupported doc_id: {doc_id}')

        dataset_version, archive_number, _ = parts[:3]
        return f'{dataset_version}-{archive_number}'

    @staticmethod
    def get_page_id(doc_id: str, page: int):
        return f'{doc_id}-page-{page}'

    @staticmethod
    def get_content_hash(doc_id: str):
        parts = doc_id.split('-')
        content_hash = next((part for part in parts if len(part) == 64), None)
        return content_hash

    @classmethod
    def get_pandas_pages_df(cls, df, feature_group_version: str, doc_id_column: str, document_column: str,
                            get_docstore_resource_bytes: Callable[..., bytes], get_document_processing_result_infos: Callable, max_workers: int = 10):
        from concurrent.futures import ThreadPoolExecutor

        import numpy as np
        import pandas as pd

        chunk_size = 10 * 1024 * 1024

        def is_valid_config(x):
            if not isinstance(x, dict):
                return False
            if cls.DOCUMENT_PROCESSING_CONFIG not in x:
                return False
            if x[cls.DOCUMENT_PROCESSING_CONFIG] is None:
                return False
            if x[cls.DOCUMENT_PROCESSING_CONFIG] == {}:
                return False
            # if all keys are None, return False
            if all(v is None for v in x[cls.DOCUMENT_PROCESSING_CONFIG].values()):
                return False
            return True

        pages_df_with_config = None
        df_with_config = df[df[document_column].apply(is_valid_config)]
        df = df[~df[doc_id_column].isin(df_with_config[doc_id_column])]

        if len(df_with_config) > 0:
            doc_ids = df_with_config[doc_id_column].values
            content_hash_to_doc_id = {cls.get_content_hash(
                doc_id): doc_id for doc_id in doc_ids}
            content_hash_list = list(content_hash_to_doc_id.keys())

            unique_document_processing_configs = set(df_with_config[document_column].apply(
                lambda x: str(x.get(cls.DOCUMENT_PROCESSING_CONFIG))))
            if len(unique_document_processing_configs) > 1:
                raise ValueError(
                    'Loading documents with different document processing configs is not supported yet. Please make sure all rows have the same document processing config.')

            sample_page_infos = df_with_config.iloc[0][document_column]
            document_processing_config = sample_page_infos[cls.DOCUMENT_PROCESSING_CONFIG]
            document_processing_version = sample_page_infos.get(
                cls.DOCUMENT_PROCESSING_VERSION)

            page_offsets_and_zip_location_list = get_document_processing_result_infos(
                content_hash_list, document_processing_config, document_processing_version) or []

            zip_location_to_offsets = {}
            for row in page_offsets_and_zip_location_list:
                zip_location_to_offsets.setdefault(
                    row['result_zip_path'], []).append(row)

            pages_list = []
            for result_zip_path, page_offsets in zip_location_to_offsets.items():
                zip_bytes = get_docstore_resource_bytes(
                    feature_group_version, cls.PAGE_DATA, result_zip_path=result_zip_path)
                for row in page_offsets:
                    start_offset, file_size = row['start_offset'], row['file_size']
                    page_content = zip_bytes[start_offset:start_offset + file_size]
                    page_data = json.loads(page_content.decode('utf-8'))
                    pages_list.append((row['content_hash'], page_data))

            json_pages_list = [{**(page or {}), doc_id_column: content_hash_to_doc_id[content_hash]}
                               for content_hash, page in pages_list]
            pages_df_with_config = pd.DataFrame(json_pages_list)
            pages_df_with_config = pages_df_with_config.replace({np.nan: None})

        df = df.drop_duplicates([doc_id_column])
        group_by_archive = df.groupby(
            df[doc_id_column].apply(lambda x: cls.get_archive_id(x)))

        def load_page(archive_bytes: bytes, offset: int, size: int):
            page_bytes = archive_bytes[offset:offset + size]
            return json.loads(page_bytes.decode('utf-8'))

        def download_archive_chunk(offset: int, size: int, archive_id: str):
            return get_docstore_resource_bytes(feature_group_version, cls.PAGE_DATA, archive_id, offset=offset, size=size)

        pages_list = []

        for archive_id, archive_group_df in group_by_archive:
            load_pages_args = []
            min_offset = None
            max_offset = None
            pages_metadata_bytes = get_docstore_resource_bytes(
                feature_group_version, cls.PAGES_ZIP_METADATA, archive_id)
            pages_metadata = json.loads(pages_metadata_bytes.decode('utf-8'))

            for doc_id, pages_ref in archive_group_df[[doc_id_column, document_column]].values:
                if not pages_ref:
                    continue
                for page in range(pages_ref[cls.FIRST_PAGE], pages_ref[cls.LAST_PAGE] + 1):
                    page_id = cls.get_page_id(doc_id, page)
                    offset, size = pages_metadata[page_id]
                    load_pages_args.append((offset, size))
                    if min_offset is None:
                        min_offset = offset
                    if max_offset is None:
                        max_offset = offset + size
                    min_offset = min(min_offset, offset)
                    max_offset = max(max_offset, offset + size)

            chunk_args = [(offset, chunk_size, archive_id)
                          for offset in range(min_offset, max_offset, chunk_size)]

            with ThreadPoolExecutor(max_workers) as executor:
                archive_chunks = executor.map(
                    lambda args: download_archive_chunk(*args), chunk_args)

            archive_bytes = b''.join(archive_chunks)

            for offset, size in load_pages_args:
                pages_list.append(
                    load_page(archive_bytes, offset - min_offset, size))

        pages_df = pd.DataFrame(pages_list)
        pages_df = pages_df.replace({np.nan: None})

        if pages_df_with_config is not None:
            pages_df = pd.concat([pages_df, pages_df_with_config])

        return pages_df

    @classmethod
    def get_pandas_documents_df(cls, df, feature_group_version: str, doc_id_column: str, document_column: str,
                                get_docstore_resource_bytes: Callable, get_document_processing_result_infos: Callable, max_workers: int = 10):
        pages_ref_column = f'__{document_column}'
        original_columns = df.columns

        # Re-name page_ids column so that the generated document column does not have the same name
        df = df.rename(columns={document_column: pages_ref_column})
        pages_df = cls.get_pandas_pages_df(df, feature_group_version, doc_id_column, pages_ref_column,
                                           get_docstore_resource_bytes, get_document_processing_result_infos, max_workers)

        # pages_df will have "doc_id" as column name which can be different from doc_id_column
        pages_df = pages_df.rename(columns={cls.DOC_ID: doc_id_column})

        def tokens_to_pages(tokens: List[dict]) -> List[str]:
            result = []
            if not tokens:
                return result
            for _, page_tokens in groupby(sorted(tokens, key=lambda t: (t[cls.PAGE], t[cls.BLOCK], t[cls.LINE])), key=lambda t: t[cls.PAGE]):
                blocks = []
                for _, block_tokens in groupby(page_tokens, key=lambda t: t[cls.BLOCK]):
                    lines = []
                    for _, line_tokens in groupby(block_tokens, key=lambda t: t[cls.LINE]):
                        lines += [' '.join(t[cls.CONTENT]
                                           for t in line_tokens), '\n']
                    if lines:
                        lines.pop()
                        blocks += [''.join(lines)]
                if blocks:
                    result.append('\n'.join(blocks))
            return result

        def combine_page_texts(pages_text: List[str]) -> str:
            return '\n'.join(pages_text)

        def tokens_to_text(tokens: List[dict]) -> str:
            return combine_page_texts(tokens_to_pages(tokens))

        # Convert column with tokens per page (list of list) to column with Document format:
        # {TOKENS: [list of tokens in document], PAGES: [list of pages in document]}.
        # No need to sort as page_df is already sorted.
        def combine_doc_info(group):
            page_infos = group.to_dict(orient='records')
            document_data = {
                cls.METADATA: [{cls.HEIGHT: page.get(cls.HEIGHT, None), cls.WIDTH: page.get(cls.WIDTH, None), cls.PAGE: page_no, cls.MARKDOWN_FEATURES: page.get(cls.MARKDOWN_FEATURES)}
                               for page_no, page in enumerate(page_infos)],
                cls.TOKENS: [token for page in page_infos for token in page.get(cls.TOKENS) or []],
                # default to embedded text
                cls.PAGES: [page.get(cls.PAGE_TEXT) or '' for page in page_infos],
                **({cls.DOC_ID: page_infos[0][cls.DOC_ID]} if cls.DOC_ID in page_infos[0] else {}),
            }
            document_data[cls.EMBEDDED_TEXT] = combine_page_texts(info.get(
                cls.EMBEDDED_TEXT) or info.get(cls.PAGE_TEXT) or '' for info in page_infos)
            page_texts = None
            for k in [cls.MULTI_MODE_OCR_TEXT, cls.PAGE_MARKDOWN, cls.PAGE_LLM_OCR, cls.PAGE_TABLE_TEXT]:
                if page_infos[0].get(k) and not document_data.get(cls.PAGE_MARKDOWN):
                    document_data[cls.PAGE_MARKDOWN] = page_texts = [
                        page.get(k, '') for page in page_infos]
                    break
            if not page_texts and page_infos[0].get(cls.EXTRACTED_TEXT):
                page_texts = [page.get(cls.EXTRACTED_TEXT, '')
                              for page in page_infos]
            elif not page_texts and page_infos[0].get(cls.TOKENS):
                page_texts = [tokens_to_text(pg[cls.TOKENS])
                              for pg in page_infos]
            if page_texts:
                document_data[cls.PAGES] = page_texts
            page_texts = page_texts or ['' for _ in page_infos]
            document_data[cls.EXTRACTED_TEXT] = combine_page_texts(page_texts)
            return document_data

        doc_infos = pages_df.groupby(doc_id_column).apply(
            combine_doc_info).reset_index(name=document_column)
        document_df = df.merge(doc_infos, on=doc_id_column, how='left')
        document_df = document_df[original_columns]
        return document_df


def try_abacus_internal_copy(src_suffix, dst_local, raise_exception=True):
    """ Retuns true if the file was copied, false otherwise"""
    safe_to_do_internal_copy = os.environ.get(
        'SAFE_TO_DO_INTERNAL_COPY') or 'true'
    if safe_to_do_internal_copy == 'true' and os.environ.get('ABACUS_GENERATED_ARTIFACTS_DIR') and importlib.util.find_spec('abacus_internal') and importlib.util.find_spec('abacus_internal.cloud_copy'):
        try:
            from abacus_internal.cloud_copy import copy_cloud_file_to_local
            copy_cloud_file_to_local(src_suffix, dst_local)
        except Exception:
            if raise_exception:
                raise Exception(
                    'Something went wrong while executing in Abacus environment. Please contact support')
        return True
    return False



from .return_class import AbstractApiClass


class ChatllmProjectPermissions(AbstractApiClass):
    """
        ChatLLM Project Permissions

        Args:
            client (ApiClient): An authenticated API Client instance
            chatllmProjectId (id): The ID of the chatllm project.
            accessLevel (str): The access level of the chatllm project.
            userPermissions (list): List of tuples containing (user_id, permission).
            userGroupPermissions (list): List of tuples containing (user_group_id, permission).
    """

    def __init__(self, client, chatllmProjectId=None, accessLevel=None, userPermissions=None, userGroupPermissions=None):
        super().__init__(client, None)
        self.chatllm_project_id = chatllmProjectId
        self.access_level = accessLevel
        self.user_permissions = userPermissions
        self.user_group_permissions = userGroupPermissions
        self.deprecated_keys = {}

    def __repr__(self):
        repr_dict = {f'chatllm_project_id': repr(self.chatllm_project_id), f'access_level': repr(
            self.access_level), f'user_permissions': repr(self.user_permissions), f'user_group_permissions': repr(self.user_group_permissions)}
        class_name = "ChatllmProjectPermissions"
        repr_str = ',\n  '.join([f'{key}={value}' for key, value in repr_dict.items(
        ) if getattr(self, key, None) is not None and key not in self.deprecated_keys])
        return f"{class_name}({repr_str})"

    def to_dict(self):
        """
        Get a dict representation of the parameters in this class

        Returns:
            dict: The dict value representation of the class parameters
        """
        resp = {'chatllm_project_id': self.chatllm_project_id, 'access_level': self.access_level,
                'user_permissions': self.user_permissions, 'user_group_permissions': self.user_group_permissions}
        return {key: value for key, value in resp.items() if value is not None and key not in self.deprecated_keys}
